@@ -1,6 +1,7 @@
 import logging
 
 from collections import namedtuple
+from datetime import datetime
 
 from django.db import transaction
 from django.db.models import Max, Q
@@ -64,10 +65,10 @@ def pre_migrate_all_content(plan):
             premigrate_hook = None
             if content_model.pulp2.TYPE_ID in plugin.migrator.premigrate_hook:
                 premigrate_hook = plugin.migrator.premigrate_hook[content_model.pulp2.TYPE_ID]
-            pre_migrate_content(content_model, mutable_type, lazy_type, premigrate_hook)
+            pre_migrate_content_type(content_model, mutable_type, lazy_type, premigrate_hook)
 
 
-def pre_migrate_content(content_model, mutable_type, lazy_type, premigrate_hook):
+def pre_migrate_content_type(content_model, mutable_type, lazy_type, premigrate_hook):
     """
     A coroutine to pre-migrate Pulp 2 content, including all details for on_demand content.
 
@@ -76,9 +77,10 @@ def pre_migrate_content(content_model, mutable_type, lazy_type, premigrate_hook)
         mutable_type: Boolean that indicates whether the content type is mutable.
     """
     batch_size = 500
-    content_type = content_model.pulp2.TYPE_ID
     pulp2content = []
     pulp2mutatedcontent = []
+    content_type = content_model.pulp2.TYPE_ID
+    set_pulp2_repo = content_model.pulp_2to3_detail.set_pulp2_repo
 
     # the latest timestamp we have in the migration tool Pulp2Content table for this content type
     content_qs = Pulp2Content.objects.filter(pulp2_content_type_id=content_type)
@@ -86,19 +88,13 @@ def pre_migrate_content(content_model, mutable_type, lazy_type, premigrate_hook)
     _logger.debug('The latest migrated {type} content has {timestamp} timestamp.'.format(
         type=content_type,
         timestamp=last_updated))
-    set_pulp2_repo = content_model.pulp_2to3_detail.set_pulp2_repo
 
     query_args = {}
-
-    if not set_pulp2_repo:
-        # query only newly created/updated items
-        query_args["_last_updated__gte"] = last_updated
-
     if premigrate_hook:
         pulp2_content_ids = premigrate_hook()
         query_args["id__in"] = pulp2_content_ids
 
-    mongo_content_qs = content_model.pulp2.objects(**query_args)
+    mongo_content_qs = content_model.pulp2.objects(_last_updated__gte=last_updated, **query_args)
     total_content = mongo_content_qs.count()
     _logger.debug('Total count for {type} content to migrate: {total}'.format(
         type=content_type,
@@ -119,9 +115,16 @@ def pre_migrate_content(content_model, mutable_type, lazy_type, premigrate_hook)
     existing_count = 0
 
     if mutable_type:
-        pulp2_content_ids = [
-            c['_id'] for c in mongo_content_qs.only('id').no_cache().as_pymongo()
-        ]
+        pulp2_content_ids = []
+
+        for c in mongo_content_qs.only('id', '_last_updated').no_cache().as_pymongo():
+            if c['_last_updated'] == last_updated:
+                if Pulp2Content.objects.filter(
+                        pulp2_last_updated=last_updated, pulp2_id=c['_id']).exists():
+                    continue
+
+                pulp2_content_ids.append(c['_id'])
+
         # This is a mutable content type. Query for the existing pulp2content.
         # If any was found, it means that the migrated content is older than the incoming.
         # Delete outdated migrated pulp2content and create a new pulp2content
@@ -164,12 +167,14 @@ def pre_migrate_content(content_model, mutable_type, lazy_type, premigrate_hook)
                 'pulp2_repository'
             )
             for relation in content_relations.iterator():
-                item = Pulp2Content(pulp2_id=record.id,
-                                    pulp2_content_type_id=record._content_type_id,
-                                    pulp2_last_updated=record._last_updated,
-                                    pulp2_storage_path=record._storage_path,
-                                    downloaded=downloaded,
-                                    pulp2_repo=relation.pulp2_repository)
+                item = Pulp2Content(
+                    pulp2_id=record.id,
+                    pulp2_content_type_id=record._content_type_id,
+                    pulp2_last_updated=record._last_updated,
+                    pulp2_storage_path=record._storage_path,
+                    downloaded=downloaded,
+                    pulp2_repo=relation.pulp2_repository
+                )
                 _logger.debug(
                     'Add content item to the list to migrate: {item}'.format(item=item))
                 pulp2content.append(item)
@@ -181,11 +186,13 @@ def pre_migrate_content(content_model, mutable_type, lazy_type, premigrate_hook)
             pulp2content_pb.total -= 1
             pulp2detail_pb.total -= 1
         else:
-            item = Pulp2Content(pulp2_id=record.id,
-                                pulp2_content_type_id=record._content_type_id,
-                                pulp2_last_updated=record._last_updated,
-                                pulp2_storage_path=record._storage_path,
-                                downloaded=downloaded)
+            item = Pulp2Content(
+                pulp2_id=record.id,
+                pulp2_content_type_id=record._content_type_id,
+                pulp2_last_updated=record._last_updated,
+                pulp2_storage_path=record._storage_path,
+                downloaded=downloaded
+            )
             _logger.debug('Add content item to the list to migrate: {item}'.format(item=item))
             pulp2content.append(item)
 
@@ -193,8 +200,9 @@ def pre_migrate_content(content_model, mutable_type, lazy_type, premigrate_hook)
         # anything in the pulp2content to be saved
         save_batch = pulp2content and (len(pulp2content) >= batch_size or i == total_content - 1)
         if save_batch:
-            _logger.debug('Bulk save for generic content info, saved so far: {index}'.format(
-                index=i + 1))
+            _logger.debug(
+                'Bulk save for generic content info, saved so far: {index}'.format(index=i + 1)
+            )
             pulp2content_batch = Pulp2Content.objects.bulk_create(pulp2content,
                                                                   ignore_conflicts=True)
             content_saved = len(pulp2content_batch) - existing_count
@@ -209,6 +217,61 @@ def pre_migrate_content(content_model, mutable_type, lazy_type, premigrate_hook)
             pulp2content.clear()
             existing_count = 0
 
+    # If a content type needs to be associated per-repo, we need to make sure that the
+    # existing content hasn't been associated with a new repo since our last migration,
+    # and if so, we need to go back and create a Pulp2Content for these new relations.
+    if set_pulp2_repo and last_updated:
+        # last_updated is a unix timestamp that sometimes defaults to 0 (if there is no
+        # existing content). If there's no existing content we can just skip this entire
+        # process; otherwise we need to convert it to use in our Django query.
+        last_updated = datetime.utcfromtimestamp(last_updated)
+
+        # This content requires to set pulp 2 repo. E.g. for errata, because 1 pulp2
+        # content unit is converted into N pulp3 content units and repo_id is the only
+        # way to have unique records for those.
+        content_relations = Pulp2RepoContent.objects.filter(
+            pulp2_content_type_id=record._content_type_id,
+            pulp2_repository__not_in_plan=False,
+            pulp2_created__gte=last_updated
+        ).select_related(
+            'pulp2_repository'
+        ).only(
+            'pulp2_repository'
+        )
+
+        mongo_content_qs = content_model.pulp2.objects(
+            id__in=content_relations.values_list('pulp2_unit_id', flat=True))
+        pulp2_content_by_id = {
+            record.id: record for record in mongo_content_qs.only(*mongo_fields).no_cache()
+        }
+
+        for relation in content_relations:
+            record = pulp2_content_by_id[relation.pulp2_unit_id]
+            downloaded = record.downloaded if hasattr(record, 'downloaded') else False
+            item = Pulp2Content(
+                pulp2_id=record.id,
+                pulp2_content_type_id=record._content_type_id,
+                pulp2_last_updated=record._last_updated,
+                pulp2_storage_path=record._storage_path,
+                downloaded=downloaded,
+                pulp2_repo=relation.pulp2_repository
+            )
+            _logger.debug(
+                'Add content item to the list to migrate: {item}'.format(item=item))
+            pulp2content.append(item)
+            pulp2content_pb.total += 1
+            pulp2detail_pb.total += 1
+
+        pulp2content_batch = Pulp2Content.objects.bulk_create(pulp2content,
+                                                              ignore_conflicts=True)
+        pulp2content_pb.done += len(pulp2content_batch)
+        pulp2content_pb.save()
+
+        content_model.pulp_2to3_detail.pre_migrate_content_detail(pulp2content_batch)
+
+        pulp2detail_pb.done += len(pulp2content_batch)
+        pulp2detail_pb.save()
+
     pulp2content_pb.save()
     pulp2detail_pb.save()
 
@@ -246,8 +309,7 @@ def pre_migrate_lazycatalog(content_type):
     pulp2lazycatalog = []
 
     mongo_lce_qs = LazyCatalogEntry.objects(unit_type_id=content_type)
-    total_lce = mongo_lce_qs.count()
-    for i, lce in enumerate(mongo_lce_qs.batch_size(batch_size).as_pymongo().no_cache()):
+    for lce in mongo_lce_qs.batch_size(batch_size).as_pymongo().no_cache():
         item = Pulp2LazyCatalog(pulp2_importer_id=lce['importer_id'],
                                 pulp2_unit_id=lce['unit_id'],
                                 pulp2_content_type_id=lce['unit_type_id'],
@@ -257,10 +319,11 @@ def pre_migrate_lazycatalog(content_type):
                                 is_migrated=False)
         pulp2lazycatalog.append(item)
 
-        save_batch = (i and not (i + 1) % batch_size or i == total_lce - 1)
-        if save_batch:
+        if len(pulp2lazycatalog) > batch_size:
             Pulp2LazyCatalog.objects.bulk_create(pulp2lazycatalog, ignore_conflicts=True)
             pulp2lazycatalog.clear()
+    else:
+        Pulp2LazyCatalog.objects.bulk_create(pulp2lazycatalog, ignore_conflicts=True)
 
 
 def pre_migrate_all_without_content(plan, type_to_repo_ids, repo_id_to_type):
@@ -319,8 +382,6 @@ def pre_migrate_all_without_content(plan, type_to_repo_ids, repo_id_to_type):
                     if not repos or repos and distributors_repos:
                         pre_migrate_distributor(
                             repo_id, distributors_repos, distributor_migrators, repo)
-                    if repo:
-                        pre_migrate_repocontent(repo)
                 pb.increment()
 
 
@@ -354,14 +415,18 @@ def pre_migrate_repo(record, repo_id_to_type):
         # if it was marked as such because it was not present in the migration plan
         repo.not_in_plan = False
         # check if there were any changes since last time
-        if last_unit_added != repo.pulp2_last_unit_added or \
-           last_unit_removed != repo.pulp2_last_unit_removed:
+        is_changed = (last_unit_added != repo.pulp2_last_unit_added
+                      or last_unit_removed != repo.pulp2_last_unit_removed)
+        if is_changed:
             repo.pulp2_last_unit_added = last_unit_added
             repo.last_unit_removed = last_unit_removed
             repo.pulp2_description = record.description
             repo.is_migrated = False
         repo.save()
 
+    if created or is_changed:
+        pre_migrate_repocontent(repo)
+
     return repo
 
 
@@ -528,19 +593,24 @@ def pre_migrate_repocontent(repo):
         return
 
     repocontent = []
-    for repocontent_data in mongo_repocontent_qs.\
-            only('unit_id', 'unit_type_id').as_pymongo().no_cache():
-        item = Pulp2RepoContent(pulp2_unit_id=repocontent_data['unit_id'],
-                                pulp2_content_type_id=repocontent_data['unit_type_id'],
-                                pulp2_repository=repo)
+    for repocontent_data in mongo_repocontent_qs.exclude('repo_id').as_pymongo().no_cache():
+        item = Pulp2RepoContent(
+            pulp2_unit_id=repocontent_data['unit_id'],
+            pulp2_content_type_id=repocontent_data['unit_type_id'],
+            pulp2_repository=repo,
+            pulp2_created=repocontent_data['created'],
+            pulp2_updated=repocontent_data['updated']
+        )
         repocontent.append(item)
 
     Pulp2RepoContent.objects.bulk_create(repocontent, batch_size=DEFAULT_BATCH_SIZE)
 
 
-def mark_removed_resources(plan, type_to_repo_ids):
+def handle_outdated_resources(plan, type_to_repo_ids):
     """
-    Marks repositories, importers which are no longer present in Pulp2.
+    Marks repositories, importers, distributors which are no longer present in Pulp2.
+
+    Delete Publications and Distributions which are no longer present in Pulp2.
 
     Args:
         plan(MigrationPlan): A Migration Plan
@@ -562,37 +632,24 @@ def mark_removed_resources(plan, type_to_repo_ids):
             str(i.id) for i in Repository.objects(mongo_repo_q).only('id'))
 
         premigrated_repos = Pulp2Repository.objects.filter(pulp2_repo_type=plugin_plan.type)
-        premigrated_repo_object_ids = set(premigrated_repos.values_list('pulp2_object_id',
-                                                                        flat=True))
+        premigrated_repo_object_ids = set(
+            premigrated_repos.values_list('pulp2_object_id', flat=True))
         removed_repo_object_ids = premigrated_repo_object_ids - mongo_repo_object_ids
 
-        removed_repos = []
-        for pulp2repo in Pulp2Repository.objects.filter(
-                pulp2_object_id__in=removed_repo_object_ids):
-            pulp2repo.not_in_plan = True
-            removed_repos.append(pulp2repo)
-
-        Pulp2Repository.objects.bulk_update(objs=removed_repos,
-                                            fields=['not_in_plan'],
-                                            batch_size=DEFAULT_BATCH_SIZE)
+        Pulp2Repository.objects.filter(
+            pulp2_object_id__in=removed_repo_object_ids).update(not_in_plan=True)
 
         # Mark importers
-        mongo_imp_object_ids = set(str(i.id) for i in Importer.objects.only('id'))
+        mongo_imp_object_ids = set(
+            str(i.id) for i in Importer.objects.only('id'))
         imp_types = plugin_plan.migrator.importer_migrators.keys()
         premigrated_imps = Pulp2Importer.objects.filter(pulp2_type_id__in=imp_types)
-        premigrated_imp_object_ids = set(premigrated_imps.values_list('pulp2_object_id',
-                                                                      flat=True))
+        premigrated_imp_object_ids = set(
+            premigrated_imps.values_list('pulp2_object_id', flat=True))
         removed_imp_object_ids = premigrated_imp_object_ids - mongo_imp_object_ids
 
-        removed_imps = []
-        for pulp2importer in Pulp2Importer.objects.filter(
-                pulp2_object_id__in=removed_imp_object_ids):
-            pulp2importer.not_in_plan = True
-            removed_imps.append(pulp2importer)
-
-        Pulp2Importer.objects.bulk_update(objs=removed_imps,
-                                          fields=['not_in_plan'],
-                                          batch_size=DEFAULT_BATCH_SIZE)
+        Pulp2Importer.objects.filter(
+            pulp2_object_id__in=removed_imp_object_ids).update(not_in_plan=True)
 
         # Mark distributors
         mongo_dist_object_ids = set(str(i.id) for i in Distributor.objects.only('id'))
@@ -602,58 +659,47 @@ def mark_removed_resources(plan, type_to_repo_ids):
                                                                         flat=True))
         removed_dist_object_ids = premigrated_dist_object_ids - mongo_dist_object_ids
 
-        removed_dists = []
-        for pulp2dist in Pulp2Distributor.objects.filter(
-                pulp2_object_id__in=removed_dist_object_ids):
-            pulp2dist.not_in_plan = True
-            removed_dists.append(pulp2dist)
+        Pulp2Distributor.objects.filter(
+            pulp2_object_id__in=removed_dist_object_ids).update(not_in_plan=True)
 
-        Pulp2Distributor.objects.bulk_update(objs=removed_dists,
-                                             fields=['not_in_plan'],
-                                             batch_size=DEFAULT_BATCH_SIZE)
+    # Delete old Publications/Distributions which are no longer present in Pulp2.
 
+    # It's critical to remove Distributions to avoid base_path overlap.
+    # It make the migration logic easier if we remove old Publications as well.
 
-def delete_old_resources(plan):
-    """
-    Delete old Publications/Distributions which are no longer present in Pulp2.
+    # Delete criteria:
+    #     - pulp2distributor is no longer in plan
+    #     - pulp2repository content changed (repo.is_migrated=False) or it is no longer in plan
 
-    It's critical to remove Distributions to avoid base_path overlap.
-    It make the migration logic easier if we remove old Publications as well.
-
-    Delete criteria:
-        - pulp2distributor is no longer in plan
-        - pulp2repository content changed (repo.is_migrated=False) or it is no longer in plan
-
-    Args:
-        plan(MigrationPlan): A Migration Plan
-
-    """
     repos_with_old_distributions_qs = Pulp2Repository.objects.filter(
-        Q(is_migrated=False) | Q(not_in_plan=True))
+        Q(is_migrated=False) | Q(not_in_plan=True)
+    )
 
     old_dist_query = Q(pulp3_distribution__isnull=False) | Q(pulp3_publication__isnull=False)
     old_dist_query &= Q(pulp2_repository__in=repos_with_old_distributions_qs) | Q(not_in_plan=True)
 
     with transaction.atomic():
         pulp2distributors_with_old_distributions_qs = Pulp2Distributor.objects.filter(
-            old_dist_query)
-        pubs_to_delete = set()
-        dists_to_delete = []
-        for pulp2distributor in pulp2distributors_with_old_distributions_qs:
-            if pulp2distributor.is_migrated:
-                pulp2distributor.is_migrated = False
-                pulp2distributor.save()
-            if pulp2distributor.pulp3_publication:
-                # check if publication is shared by multiple distributions
-                # on the corresponding distributor flip the flag to false so the affected
-                # distribution will be updated with the new publication
-                pulp2dists = pulp2distributor.pulp3_publication.pulp2distributor_set.all()
-                for dist in pulp2dists:
-                    if dist.is_migrated:
-                        dist.is_migrated = False
-                        dist.save()
-                pubs_to_delete.add(pulp2distributor.pulp3_publication.pk)
-            if pulp2distributor.pulp3_distribution:
-                dists_to_delete.append(pulp2distributor.pulp3_distribution.pk)
-        Publication.objects.filter(pk__in=pubs_to_delete).delete()
-        BaseDistribution.objects.filter(pk__in=dists_to_delete).delete()
+            old_dist_query
+        )
+
+        pulp2distributors_with_old_distributions_qs.update(
+            is_migrated=False
+        )
+
+        # If publication is shared by multiple distributions, on the corresponding distributors
+        # flip the flag to false so the affected distributions will be updated with the new
+        # publication
+        Pulp2Distributor.objects.filter(
+            pulp3_publication__in=Publication.objects.filter(
+                pulp2distributor__in=pulp2distributors_with_old_distributions_qs
+            )
+        ).update(is_migrated=False)
+
+        # Delete outdated publications
+        Publication.objects.filter(
+            pulp2distributor__in=pulp2distributors_with_old_distributions_qs).delete()
+
+        # Delete outdated distributions
+        BaseDistribution.objects.filter(
+            pulp2distributor__in=pulp2distributors_with_old_distributions_qs).delete()
